/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include "util.h"
#include "getopt.h"
#include "globalregistry.h"
#include "configfile.h"
#include "kis_alertclient.h"

// Some system defaults
#define SYSCONF_LOC "/usr/local/etc/"
const char *config_base = "kismet_alertlogger.conf";

char *outfname = NULL;

void WriteHTMLTable(FILE *outf, vector<KismetAlertClient::alert_info *> avec,
					GlobalRegistry *globalreg) {

	fprintf(outf, "<table border=\"1\" class=\"kisal\">\n");
	fprintf(outf, "<tr><th class=\"kisal\">Time</th>"
			"<th class=\"kisal\">Alert</th>"
			"<th class=\"kisal\">BSSID</th>"
			"<th class=\"kisal\">Source</th>"
			"<th class=\"kisal\">Dest</th>"
			"<th class=\"kisal\">Other</th>"
			"<th class=\"kisal\">Ch</th>"
			"<th class=\"kisal\">Info</th></tr>\n");
	
	if (avec.size() == 0) {
		fprintf(outf, "<tr><td colspan=\"8\" class=\"kisal\">"
				"<i>Currently no alerts</i></td></tr>\n");
	}
	
	for (unsigned int x = 0; x < avec.size(); x++) {
		fprintf(outf, "<tr>\n");
		fprintf(outf, "<td class=\"kisal\">%s</td>\n", 
				ctime((const time_t *) &(avec[x]->alert_ts.tv_sec)));
		fprintf(outf, "<td class=\"kisal\">%s</td>\n", avec[x]->header.c_str());
		fprintf(outf, "<td class=\"kisal\">%s</td>"
				"<td class=\"kisal\">%s</td>"
				"<td class=\"kisal\">%s</td><td class=\"kisal\">%s</td>\n",
				avec[x]->bssid.Mac2String().c_str(),
				avec[x]->source.Mac2String().c_str(),
				avec[x]->dest.Mac2String().c_str(),
				avec[x]->other.Mac2String().c_str());
		fprintf(outf, "<td class=\"kisal\">%d</td>\n", avec[x]->channel);
		fprintf(outf, "<td class=\"kisal\">%s</td>\n", avec[x]->text.c_str());
	}

	fprintf(outf, "</table>");
}

// Core timer event that writes alerts out
int AlertWriterEvent(Timetracker::timer_event *evt, void *parm,
					 GlobalRegistry *globalreg) {
	vector<KismetAlertClient::alert_info *> avec =
		globalreg->kisalertclient->FetchAlerts();
	
	FILE *outf = NULL;
	char err[1024];

	if ((outf = fopen(outfname, "w")) == NULL) {
		snprintf(err, 1024, "Unable to open '%s' for writing", outfname);
		globalreg->messagebus->InjectMessage(err, MSGFLAG_FATAL);
		globalreg->fatal_condition = 1;
		return 0;
	}

	WriteHTMLTable(outf, avec, globalreg);

	if (outf)
		fclose(outf);

	sync();

	globalreg->messagebus->InjectMessage("Log file written.", MSGFLAG_INFO);
	
	return 1;
}

void Usage(char *argv) {
	fprintf(stderr, "%s [ -f configfile ] [ -o outputfile ]\n", argv);

	exit(1);
}

void CatchShutdown(int reason) {
	fprintf(stderr, "FATAL - Shutting down.\n");
	exit(1);
}

int main(int argc, char *argv[]) {
	char status[1024];

	static struct option long_options[] = {
		{ "config-file", required_argument, 0, 'f' },
		{ "output-file", required_argument, 0, 'o' }
	};

	char *configfile = NULL;

	int option_index;

	while (1) {
		int r = getopt_long(argc, argv, "f:o:",
							long_options, &option_index);

		if (r < 0)
			break;

		switch (r) {
			case 'f':
				configfile = optarg;
				break;

			case 'o':
				outfname = strdup(optarg);
				break;

			default:
				Usage(argv[0]);
				break;
		}
	}
	
	// Build internal structures
	GlobalRegistry *globalreg = new GlobalRegistry;
	
	// Create the messagebus and clients
	globalreg->messagebus = new MessageBus;
	StdoutMessageClient *stdmsgcli = new StdoutMessageClient(globalreg);
	globalreg->messagebus->RegisterClient(stdmsgcli, MSGFLAG_ALL);

	// Allocate a timetracker
	globalreg->timetracker = new Timetracker(globalreg);

	// Config file has to be opened before kis_alertclient is created or it 
	// will be very sad.
	int freeconf = 0;
	if (configfile == NULL) {
		configfile = (char *) malloc(1024 * sizeof(char));
		snprintf(configfile, 1024, "%s/%s", SYSCONF_LOC, config_base);
		freeconf = 1;
	}

	ConfigFile *conf = new ConfigFile;

	// Parse the config file
	if (conf->ParseConfig(configfile) < 0) {
		exit(1);
	}

	// Grab the output file name
	if (conf->FetchOpt("alertfile") != "") {
		outfname = strdup(conf->ExpandLogPath(conf->FetchOpt("alertfile"), 
											  "", "", 0, 1).c_str());
	} else if (outfname == NULL) {
		globalreg->messagebus->InjectMessage("No output file given on command line "
											 "and none in config file.", 
											 MSGFLAG_FATAL);
		globalreg->fatal_condition = 1;
		CatchShutdown(-1);
	}

	// Assign it to the global registry
	globalreg->kismetui_config = conf;

	// Create the network client
	KismetAlertClient *kis_alc = new KismetAlertClient(globalreg);
	globalreg->kisalertclient = kis_alc;


	// Register a timer event for our alert printer
	globalreg->timetracker->RegisterTimer(SERVER_TIMESLICES_SEC * 10,
										  NULL, 1, &AlertWriterEvent, NULL);

	while (1) {
		int max_fd = 0;
		fd_set rset, wset;
		struct timeval tm;

		FD_ZERO(&rset);
		FD_ZERO(&wset);

		max_fd = globalreg->kisalertclient->MergeSet(rset, wset, max_fd,
													 &rset, &wset);

		tm.tv_sec = 0;
		tm.tv_usec = 100000;

		if (select(max_fd + 1, &rset, &wset, NULL, &tm) < 0) {
			if (errno != EINTR) {
				snprintf(status, STATUS_MAX, 
						 "select() error (%s)", strerror(errno));
				globalreg->messagebus->InjectMessage(status, MSGFLAG_FATAL);
				CatchShutdown(-1);
			}

		}

		// Only bail on fully fatal errors
		if (globalreg->kisalertclient->Poll(rset, wset) < 0 &&
			globalreg->fatal_condition)
			CatchShutdown(-1);

		globalreg->timetracker->Tick();

	}
	
}

