/*
    This file is part of Kismet

    Kismet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Kismet is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Kismet; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include "util.h"
#include "configfile.h"
#include "timetracker.h"
#include "kis_alertclient.h"

// Reconnect trigger
int KisAlCliReconEvent(Timetracker::timer_event *evt, void *parm,
					   GlobalRegistry *globalreg) {
	KismetAlertClient *alc = (KismetAlertClient *) parm;

	alc->Reconnect();

	return 1;
}

KismetAlertClient::KismetAlertClient() {
    fprintf(stderr, "*** KismetAlertClient called with no global registry reference\n");
    globalreg = NULL;
    tcpcli = NULL;
}

KismetAlertClient::KismetAlertClient(GlobalRegistry *in_globalreg) : 
	ClientFramework(in_globalreg) {
    // The only connection type we support right now is TCP so just make it manually
    // this could be parsed from the config file when we support other transport
    // layers
    tcpcli = new TcpClient(globalreg);

    // Attach it to ourselves and opposite
    RegisterNetworkClient(tcpcli);
    tcpcli->RegisterClientFramework(this);

    // Initialize all our stuff
    cmdid = 1;

    reconnect_attempt = -1;

    server_identified = 0;
    
    maxalerts = 500;

    // Fill in the protocol default field lists
    protocol_default_map["ALERT"] = "sec,usec,bssid,source,dest,other,channel,"
		"header,text";
	protocol_map["ALERT"] = 1;
    protocol_default_map["KISMET"] = "version,starttime,servername,timestamp,"
        "channelhop,newversion";
	protocol_map["KISMET"] = 1;


    // Parse the config file and enable the tcpclient
    char temphost[128];
    if (sscanf(globalreg->kismetui_config->FetchOpt("host").c_str(), "%128[^:]:%d", 
               temphost, &port) != 2) {
        globalreg->messagebus->InjectMessage("Invalid Kismet host in config, "
											 "host:port required", MSGFLAG_FATAL);
        globalreg->fatal_condition = 1;
        return;
    }
    snprintf(host, MAXHOSTNAMELEN, "%s", temphost);

	if (globalreg->kismetui_config->FetchOpt("maxalerts") != "") {
		if ((sscanf(globalreg->kismetui_config->FetchOpt("maxalerts").c_str(), "%d",
					&maxalerts)) != 1) {
			globalreg->messagebus->InjectMessage("Invalid 'maxalerts' value in "
												 "config file.", MSGFLAG_FATAL);
			globalreg->fatal_condition = 1;
			return;
		}
	}

    if (globalreg->kismetui_config->FetchOpt("reconnect") == "true") {
        globalreg->messagebus->InjectMessage("Enabling reconnection to the Kismet "
											 "server if the link is lost", 
											 MSGFLAG_INFO);
        reconnect_attempt = 0;
    }

    if (tcpcli->Connect(host, port) < 0) {
        globalreg->messagebus->InjectMessage("Could not create initial connection to "
                                             "the Kismet server", MSGFLAG_ERROR);

		if (reconnect_attempt < 0) {
			globalreg->messagebus->InjectMessage("Connection failed and reconnecting "
												 "disabled.", MSGFLAG_FATAL);
			globalreg->fatal_condition = 1;

			return;
		}
		
        last_disconnect = time(0);
    } else {
        // Turn on the protocols in our map, if we have any.  This is for reconnecting
        // to a server and preserving the requested protocols
        for (map<string, int>::iterator prot = protocol_map.begin();
             prot != protocol_map.end(); ++prot) {
            EnableProtocol(prot->first);

            if (globalreg->fatal_condition)
                return;
        }
    }

	if (reconnect_attempt >= 0)
		reconid = 
			globalreg->timetracker->RegisterTimer(SERVER_TIMESLICES_SEC * 5,
												  NULL, 1, &KisAlCliReconEvent,
												  (void *) this);
	else
		reconid = -1;

    snprintf(errstr, STATUS_MAX, "Using Kismet server on %s:%d", host, port);
    globalreg->messagebus->InjectMessage(errstr, MSGFLAG_INFO);
}

KismetAlertClient::~KismetAlertClient() {
    if (tcpcli != NULL) {
        tcpcli->KillConnection();
        delete tcpcli;
    }
    if (reconid >= 0 && globalreg != NULL)
        globalreg->timetracker->RemoveTimer(reconid);
}

int KismetAlertClient::KillConnection() {
    if (tcpcli != NULL)
        tcpcli->KillConnection();

    return 1;
}

int KismetAlertClient::Shutdown() {
    if (tcpcli != NULL) {
        tcpcli->FlushRings();
        tcpcli->KillConnection();
    }

    return 1;
}

void KismetAlertClient::EnableProtocol(string in_protocol) {
    string fields;
    char data[1024];

    if (protocol_default_map.find(in_protocol) != protocol_default_map.end())
        fields = protocol_default_map[in_protocol];
    else
        fields = "*";

    // We don't care about ACK or errors, because we suck
    snprintf(data, 1024, "ENABLE %s %s\n", in_protocol.c_str(), fields.c_str());

    protocol_map[in_protocol] = 1;
    
	InjectCommand(data);
}

int KismetAlertClient::InjectCommand(string in_cmd) {
    if (tcpcli->Valid() == 0) {
		return 0;
    }

    int curid = cmdid++;
    char cmdheader[32];

    snprintf(cmdheader, 32, "!%d ", curid);
    string fullcmd = string(cmdheader) + in_cmd;

    if (tcpcli->Valid() && tcpcli->WriteData((void *) fullcmd.c_str(), 
											 fullcmd.length()) < 0 ||
        globalreg->fatal_condition) {
        last_disconnect = time(0);
        return -1;
    }

    return curid;
}

int KismetAlertClient::Reconnect() {
	if (tcpcli->Valid())
		return 1;
	
    if (tcpcli->Connect(host, port) < 0) {
        globalreg->messagebus->InjectMessage("Could not reconnect to Kismet server", 
											 MSGFLAG_ERROR);
        reconnect_attempt++;
        last_disconnect = time(0);
        return 0;
    } else {
        // Turn on the protocols in our map, if we have any.  This is for reconnecting
        // to a server and preserving the requested protocols
        for (map<string, int>::iterator prot = protocol_map.begin();
             prot != protocol_map.end(); ++prot) {
            EnableProtocol(prot->first);

            if (globalreg->fatal_condition)
                return 0;
        }
    }

	globalreg->messagebus->InjectMessage("Reconnected to the Kismet server",
										 MSGFLAG_INFO);
    
    return 1;
}

int KismetAlertClient::ParseData() {
    int len, rlen;
    char *buf;
    string strbuf;

    // Scratch variables for parsing data
    char header[65];
    char bssid_str[18];
    mac_addr bssid;

    len = netclient->FetchReadLen();
    buf = new char[len + 1];
    
    if (netclient->ReadData(buf, len, &rlen) < 0) {
        globalreg->messagebus->InjectMessage("KismetAlertClient::ParseData failed to "
											 "fetch data from the tcp connection.", 
											 MSGFLAG_ERROR);
        return -1;
    }
    buf[len] = '\0';

    // Parse without including partials, so we don't get a fragmented command 
    // out of the buffer
    vector<string> inptok = StrTokenize(buf, "\n", 0);
    delete[] buf;

    // Bail on no useful data
    if (inptok.size() < 1) {
        return 0;
    }

    for (unsigned int it = 0; it < inptok.size(); it++) {
        // No matter what we've dealt with this data block
        netclient->MarkRead(inptok[it].length() + 1);

        // Pull the header out to save time -- cheaper to parse the header and 
		// then the data than to try to parse an entire data string just to find 
		// out what protocol we are
        // 
        // Protocol parsers should be dynamic so that we can have plugins in the 
		// framework able to handle a proto, but right now thats a hassle

        if (sscanf(inptok[it].c_str(), "*%64[^:]", header) < 1) {
            continue;
        }

        // Nuke the header off the string
        inptok[it].erase(0, (size_t) strlen(header) + 3);

        if (!strncmp(header, "TERMINATE", 64)) {
            snprintf(errstr, STATUS_MAX, "Kismet server terminated.");
            globalreg->messagebus->InjectMessage(errstr, MSGFLAG_ERROR);
           
            netclient->KillConnection();
            
            continue;
        } else if (!strncmp(header, "KISMET", 64)) {
            // Parse kismet protocol or skip
			// We don't really care much whats in here
    		int junkmajor, junkminor, junktiny;
			int start_time;
			char servername[33];
			char build[25];
			int channel_hop;
			char major[25], minor[25], tiny[25];

            if (sscanf(inptok[it].c_str(), "%d.%d.%d %d \001%32[^\001]\001 %24s %d "
                       "%24[^.].%24[^.].%24s",
                       &junkmajor, &junkminor, &junktiny, 
                       (int *) &start_time, servername, 
                       build, &channel_hop,
                       major, minor, tiny) < 7)
                continue;

            server_identified = 1;
		} else if (!strncmp(header, "ALERT", 64)) {
			// Parse the ALERT protocol
			char a_bssid_str[18];
			char a_src_str[18];
			char a_dest_str[18];
			char a_other_str[18];
			int a_ch;
			char a_hdr[128];
			char a_txt[2048];
			long int a_tvsec, a_tvusec;

			if (sscanf(inptok[it].c_str(), "%ld %ld %17s %17s %17s %17s %d %129s "
					   "\001%2047[^\001]\001",
					   &a_tvsec, &a_tvusec, a_bssid_str, a_src_str,
					   a_dest_str, a_other_str, &a_ch, a_hdr, a_txt) < 9)
				continue;

			alert_info *newal = new alert_info;

			newal->alert_ts.tv_sec = a_tvsec;
			newal->alert_ts.tv_usec = a_tvusec;
			newal->bssid = a_bssid_str;
			newal->source = a_src_str;
			newal->dest = a_dest_str;
			newal->other = a_other_str;
			newal->channel = a_ch;
			newal->header = string(a_hdr);
			newal->text = string(a_txt);

			if (alerts.size() > maxalerts) {
				delete alerts[0];
				alerts.erase(alerts.begin());
			}
			alerts.push_back(newal);
		}

    }

    return 1;
}

